// Local header include
#include "adc_interface.h"

using namespace std;

#define DEFAULT_CRYSTAL_FREQUENCY  25000000
#define ADC_SPI_ADDRESS            0x9E

struct GCTL_Byte_Indices
{
    uint8_t MODE0 = 0;         // 0:1 sample rate range
    uint8_t MODE1 = 1;
    
    uint8_t DIF0 = 2;          // 2:3 data format of serial audio
    uint8_t DIF1 = 3;
    
    uint8_t MDIV0 = 4;         // 4:5 divider for crystal clock frequency
    uint8_t MDIV1 = 5;
    
    uint8_t CLKMODE = 6;      // 6   if enabled, divides crystal clock frequency 1.5
    uint8_t CP_EN = 7;        // 7   enabled Control Port Mode
};

struct GCTL_Register
{
    uint8_t data;
    struct GCTL_Byte_Indices indices;
};

struct PDN_Byte_Indices
{
    uint8_t PDN12 = 0;        // power down channel pair 1 and 2
    uint8_t PDN34 = 1;
    uint8_t PDN56 = 2;
    uint8_t PDN78 = 3;
    uint8_t PDN_OSC = 4;      // disable internal oscillator core
    uint8_t PDN_BG = 5;       // powers down bandgap reference
    uint8_t RSVD0 = 6;
    uint8_t RSVD1 = 7;
};

struct PDN_Register
{
    uint8_t data;
    struct PDN_Byte_Indices indices;
};

struct SDEN_Byte_Indices
{
    uint8_t SDEN1 = 0;
    uint8_t SDEN2 = 1;
    uint8_t SDEN3 = 2;
    uint8_t SDEN4 = 3;
    uint8_t RSVD0 = 4;
    uint8_t RSVD1 = 5;
    uint8_t RSVD2 = 6;
    uint8_t RSVD3 = 7;
};

struct SDEN_Register
{
    uint8_t data;
    struct SDEN_Byte_Indices indices;
};

typedef struct CS5368_Registers
{
    uint8_t GCTL_addr;            // Global Mode Control Register
    struct GCTL_Register GCTL;    // bit map for the register

    uint8_t OVFLn_addr;          // Overflow Status Register
    uint8_t OVFLn;               // bit 1 = channel 1 ... bit 8 = channel 8
    
    uint8_t OVFM_addr;           // Overflow Mask Register to enable/disable interrupts
    uint8_t OVFM;                // bit 1 = channel 1 ... bit 8 = channel 8
    
    uint8_t HPFn_addr;           // High Pass Filter Register
    uint8_t HPFn;                // bit 1 = channel 1 ... bit 8 = channel 8

    uint8_t PDN_addr;             // Power Down Register
    struct PDN_Register PDN;      // bit map for the register
    
    uint8_t MUTE_addr;           // mute individual channels
    uint8_t MUTE;                // bit 1 = channel 1 ... bit 8 = channel 8
    
    uint8_t SDENn_addr;          // SDOUT enable control register
    struct SDEN_Register SDENn;   // bit map for the register
} ADC_Registers;

static ADC_Registers adc_registers;

AdcDevice::AdcDevice()
    :   crystalFrequency(DEFAULT_CRYSTAL_FREQUENCY),
        chipAddress(ADC_SPI_ADDRESS)
{}

AdcDevice::~AdcDevice() {}

void AdcDevice::initDevice()
{
    adc_registers.GCTL_addr = 0x01;
    adc_registers.OVFLn_addr = 0x02;
    adc_registers.OVFM_addr = 0x03;
    adc_registers.HPFn_addr = 0x04;
    adc_registers.PDN_addr = 0x06;
    adc_registers.MUTE_addr = 0x08;
    adc_registers.SDENn_addr = 0x0A;

    // Set Global Control Register and write to ADC via SPI
    adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.CP_EN);          // enable control port mode
    adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.CLKMODE);        // no divide by 1.5
    adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MDIV0);          // divide by 2
    adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV1);
    adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.DIF0);
    adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.DIF1);
    adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MODE0);
    adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MODE1);

    // Initialize other register values to default ADC values
    adc_registers.OVFLn = 0xFF;                   // no overflows have occurred
    adc_registers.OVFM = 0xFF;                    // overflow interrupt enabled
    adc_registers.HPFn = 0x00;                    // high pass filter enabled
    adc_registers.PDN.data = 0x00;                // everything is powered
    adc_registers.MUTE = 0x00;                    // Don't mute any channels
    adc_registers.SDENn.data = 0x00;              // SDOUT pins enabled
    
    // ADC class has no idea how information gets to its registers,
    // so it that contains data to be written to it
    printf("Initialized ADC device\n");
}

void AdcDevice::setHighPassFilter(bool set)
{
    if(set)
    {
        adc_registers.HPFn = 0x00;
    }
    else
    {
        adc_registers.HPFn = 0xFF;
    }
}

void AdcDevice::setSamplingRate(uint32_t rate)
{
    // Set GCTL bits based on sampling rate & crystal clock frequency
    double xtiClock = crystalFrequency;

    if(rate > 108000)
    {
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MODE0);
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MODE1);
    }
    else if(rate > 54000)
    {
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MODE0);
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MODE1);
        xtiClock /= 2;
    }
    else
    {
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MODE0);
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MODE1);
        xtiClock /= 4;
    }

    // Calculate avaliable rates derived from clock freq
    double max_rates[5];
    max_rates[0] = (double) xtiClock / 64;
    max_rates[1] = max_rates[0] / 1.5;
    max_rates[2] = max_rates[0] / 2;
    max_rates[3] = max_rates[0] / 3;
    max_rates[4] = max_rates[0] / 4;
    
    // Set clock dividers appropriately
    if(rate < max_rates[0]+5 && rate > max_rates[0]-5)
    {
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.CLKMODE);
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV0);           // divide by 2
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV1);
    }
    else if(rate < max_rates[1]+5 && rate > max_rates[1]-5)
    {
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.CLKMODE);
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV0);           // divide by 2
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV1);
    }
    else if(rate < max_rates[2]+5 && rate > max_rates[2]-5)
    {
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.CLKMODE);
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MDIV0);           // divide by 2
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV1);
    }
    else if(rate < max_rates[3]+5 && rate > max_rates[3]-5)
    {
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.CLKMODE);
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MDIV0);           // divide by 2
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.MDIV1);
    }
    else if(rate < max_rates[4]+5 && rate > max_rates[4]-5)
    {
        adc_registers.GCTL.data |= (0x0 << adc_registers.GCTL.indices.CLKMODE);
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MDIV0);           // divide by 2
        adc_registers.GCTL.data |= (0x1 << adc_registers.GCTL.indices.MDIV1);
    }
    else
    {
        std::cout << "Unknown sampling rate. "
                  << "Valid rates with current crystal are:" << std::endl;
        double dividers[] = {1, 1.5, 2, 3, 4};
        for(int i = 1; i <= 4; i*=2)
        {
            double mclk = crystalFrequency;
            mclk = crystalFrequency / 64 / i;
            for(int j = 0; j < 5; ++j)
            {
                rate = mclk / dividers[j];
                if(i == 1 && rate <= 216000 && rate >= 108000)
                {
                    cout << "quadruple-speed mode: " << (int)rate << " Hz" << endl;
                }
                else if(i == 2 && rate < 108000 && rate >= 54000)
                {
                    cout << "double-speed mode:    " << (int)rate << " Hz" << endl;
                }
                else if(i == 4 && rate >= 2000 && rate < 54000)
                {
                    cout << "single-speed mode:    " << (int)rate << " Hz" << endl;
                }
            }
        }
        // exit(1);
    }
}

uint8_t AdcDevice::getSpiAddress(void)
{
    return chipAddress;
}

uint8_t AdcDevice::getGCTLData(void)
{
    return adc_registers.GCTL.data;
}

uint8_t AdcDevice::getGCTLAddress(void)
{
    return adc_registers.GCTL_addr;
}