#include "audio_synthesizer.h"

// For sleep functions
#include <chrono>
#include <thread>

#define WRITE_BIT 0x0
#define READ_BIT 0x1

using namespace std;

/* Table and function used to flip a byte, defined for convenient
 * debugging purposes */
uint8_t lut[16] = {
    0x0, 0x8, 0x4, 0xC,
    0x2, 0xA, 0x6, 0xE,
    0x1, 0x9, 0x5, 0xD,
    0x3, 0xB, 0x7, 0xF };
uint8_t flip( uint8_t n )
{
   return (lut[n&0x0F] << 4) | lut[n>>4];
}

AudioSynthesizer::AudioSynthesizer() {}

AudioSynthesizer::~AudioSynthesizer() {}

void AudioSynthesizer::initDevice()
{
    // Initialize both devices
    usb_device.initDevice();

    adc_device.initDevice();
    uint8_t data_to_write = adc_device.getGCTLData();
    uint8_t address = adc_device.getGCTLAddress();
    write_SPI(data_to_write, address);
    
    adc_device.setSamplingRate(24414);
    data_to_write = adc_device.getGCTLData();
    write_SPI(data_to_write, address);
}

void AudioSynthesizer::start()
{
    // Release flip flop from reset
    usb_device.start();
}

void AudioSynthesizer::setHighPassFilter(bool set)
{
    uint8_t data = 0;
    uint8_t address = 0;
    adc_device.setHighPassFilter(set);
    write_SPI(data, address);
}

void AudioSynthesizer::write_SPI(uint8_t dataIn, uint8_t regAddress)
{
    uint8_t adcAddress = adc_device.getSpiAddress() | WRITE_BIT;
    uint8_t registerAddress = regAddress;
    uint8_t data = dataIn;                // copy data
    uint32_t clock_wait = 25;             // clock change wait time

    // 16 clock changes to write 8 bits
    uint8_t num_clock_changes = 16;
    
    // Chip address
    adcAddress = flip(adcAddress);

    PinState bit_to_write;
    for(int i = 0; i < num_clock_changes; ++i)
    {
        if(i == 0) 
        {
            // drop select when starting
            usb_device.CSn = LOW;
        }
        usb_device.CCLK = !usb_device.CCLK;     // flip clock every time

        // Update bit every time clock drops
        if(i % 2 == 0)
        {
            bit_to_write = adcAddress & 0x01 ? HIGH : LOW;
            usb_device.CDIN = bit_to_write;
            adcAddress >>= 1;
        }
        // usb_device.CDIN.write(); // write pin state to hardware
        std::this_thread::sleep_for(std::chrono::milliseconds(clock_wait));             // delay
    }

    // Register address
    registerAddress = flip(registerAddress);
    for(int i = 0; i < num_clock_changes; ++i)
    {
        usb_device.CCLK.set(!usb_device.CCLK);
        if(!usb_device.CCLK) // if clock is low
        {
            bit_to_write = registerAddress & 0x01 ? HIGH : LOW;
            usb_device.CDIN.set(bit_to_write);
            registerAddress >>= 1;
        }
        usb_device.CDIN.write();
        std::this_thread::sleep_for(std::chrono::milliseconds(clock_wait));
    }

    // Data
    data = flip(data);
    for(int i = 0; i < num_clock_changes + 2; ++i)
    {
        usb_device.CCLK.set(!usb_device.CCLK);
        if(!usb_device.CCLK && i < num_clock_changes)
        {
            bit_to_write = data & 0x01 ? HIGH : LOW;
            usb_device.CDIN.set(bit_to_write);
            data >>= 1;
        }
        if(i == num_clock_changes)
        {
            usb_device.CSn.set(HIGH);
        }
        usb_device.CDIN.write();
        std::this_thread::sleep_for(std::chrono::milliseconds(clock_wait));
    }
}

