#ifndef __BUFFER_H_
#define __BUFFER_H_

template<class T>
class Buffer 
{
    public:
        Buffer()
            :   size(0),
                head(0),
                entries(0),
                current(0),
                buffer(0)
        {}

        Buffer(uint32_t len)
            :   size(len),
                head(0),
                entries(0),
                current(0) 
        {
            buffer = new T[size];
        }

        Buffer(const Buffer &rhs) 
            :   size(rhs.size), 
                head(rhs.head), 
                entries(rhs.entries), 
                current(rhs.current)
        {
            buffer = new T[size];
            for (uint32_t i = 0; i < size; ++i) {
                buffer[i] = rhs.buffer[i];
            }
        }

        Buffer & operator=(const Buffer& rhs)
        {
            //Protect against self assignment
            if (this != rhs) 
            {
                //Create a new buffer, copy the right had data into the temp buffer
                T* new_buffer = new T[rhs.size];
                for (uint32_t i = 0; i < rhs.size; ++i) {
                    new_buffer[i] = rhs.buffer[i];
                }

                //Delete the current buffer if it exists
                if (buffer != NULL) {
                    delete[] buffer;
                }
                //Assign the buffer to the temp buffer
                buffer = new_buffer;

                //Copy other members
                size = rhs.size;
                head = rhs.head;
                entries = rhs.entries;
                current = rhs.current;

            }
            return *this;
        }

        ~Buffer() 
        {
            if (buffer != NULL) 
            {
                delete[] buffer;
            }
        }

        void setSize(uint32_t size) 
        {
            this->size = size;
            if (buffer != NULL)
            {
                delete[] buffer;
            }
            buffer = new T[size];
            head = 0;
            entries = 0;
            current = 0;
        }

        void clearAll()
        {
            head = 0;
            entries = 0;
            current = 0;
        }

        void reset()
        {
            current = 0;
        }
        
        uint32_t getEntries()
        {
            return entries;
        }
        
        /* Gets the next item until the end is reached */
        bool getNext(T& item) 
        {
            if (current == -1 || entries == 0 || 
                current == (head+entries) % size)
            {
                return false;
            }
            item = buffer[current];
            current = (current + 1) % size;
            if (current == head)
            {
                current = -1;
            }
            return true;
        }
        
        /* Retrieve first element and remove from buffer */
        bool pop(T* item) 
        {
            if (entries < 1)
            {
                return false;
            }
            *item = buffer[head];
            head = (head + 1) % size;
            entries--;
            return true;
        }
        
        /* Store the first N elements but don't remove from buffer */
        bool getN(T* buf, uint32_t n)
        {
            if (n > entries)
            {
                return false;
            }
            for (uint32_t i = 0; i < n; i++)
            {
                buf[i] = buffer[(head + i) % size];
            }
            return true;
        }
        
        /* Remove the first N elements from buffer */
        bool clearN(uint32_t n) 
        {
            if (n > entries)
            {
                return false;
            }
            head = (head + n) % size;
            entries -= n;
            return true;
        }
        /* Add an item to the end of the buffer */
        uint32_t add(T item) 
        {
            if (entries == size)
            {
                ++head;
            }
            uint32_t index = (head + entries) % size;
            buffer[index] = item;
            if (entries != size)
            {
                ++entries;
            }
            return index;
        }

        /* Add N items to the end of the buffer */
        void addN(T* buf, uint32_t n) 
        {
            for (uint32_t i = 0; i < n; i++) 
            {
                if (entries == size)
                {
                    ++head;
                }
                buffer[(head + entries) % size] = buf[i];
                if (entries != size)
                {
                    ++entries;
                }
            }
        }

        /* Overloaded indexing operator for easy access, no bound checking */
        T& operator[](uint32_t i) 
        {
            return (buffer[(head + i) % size]);
        }
    
    private:
        T* buffer;
        uint32_t size;          // max size of buffer
        uint32_t head;          // location of first element
        uint32_t entries;          // number of entries
        uint32_t current;
};
#endif //__BUFFER_H_
