// Local header include
#include "usb_interface.h"

// For sleep functions
#include <chrono>
#include <thread>

using namespace std;

#define CBUS_ALL_OUTPUTS           0xF0
#define ADC_SPI_ADDRESS            0x9E

static const uint32_t BYTES_TO_READ = 50;

FtdiDevice::FtdiDevice()
    :   ftStatus(0),
        ftHandle(0),
        numChannels(DEFAULT_NUM_CHANNELS),
        BytesReceived(0),
        RxBuffer(),
        RST(this, 0),
        CDIN(this, 1),
        CCLK(this, 2),
        CSn(this, 3),
        cbusState(CBUS_ALL_OUTPUTS),
        chipAddress(ADC_SPI_ADDRESS)
{
    for(uint32_t i = 0; i < 1024; ++i) RxBuffer[i] = 0;
    dataBuffer.setSize(265000);
}

FtdiDevice::~FtdiDevice() 
{
    closeDevice();
}

void FtdiDevice::initDevice()
{
    // initialize USB Device
    openDevice();
    resetDevice();

    // Clear Tx/Rx buffers
    clear();

    // program chip memory first
    programEEPROM();

    // Set USB pins that begin to initialize the ADC Device
    RST.set(HIGH);  // Hold ADC in reset
    CDIN.set(LOW);  // Set data line to zero
    CCLK.set(HIGH); // Set clock high
    CSn = LOW;      // Clear the flip flop and write initial states to ADC

    std::this_thread::sleep_for(std::chrono::milliseconds(20));

    RST = LOW;      // Release ADC from reset
    std::this_thread::sleep_for(std::chrono::milliseconds(20));

    printf("Initialized USB device\n");
}

void FtdiDevice::start()
{
    CSn = HIGH;
    read(5000);
}

void FtdiDevice::programEEPROM()
{
    static FT_PROGRAM_DATA programData;

    // These values are specified in ftd2xx.h starting at line 539
    programData.Signature1 = 0x00000000;
    programData.Signature2 = 0xFFFFFFFF;
    programData.Version = 5;
    programData.VendorId = 0x0403;
    programData.ProductId = 0x6014;
    
    programData.Manufacturer = "FTDI";
    programData.ManufacturerId = "FT";
    programData.Description = "UM232H";
    programData.SerialNumber = "FT000001";

    programData.MaxPower = 500;
    programData.PnP = 1;
    programData.SelfPowered = 0;
    programData.RemoteWakeup = 1;

    programData.Rev5 = 1;
    programData.IsoIn = 0;
    programData.IsoOut = 0;
    programData.PullDownEnable = 1;
    programData.SerNumEnable = 1;
    programData.SerNumEnableH = 1;
    programData.USBVersionEnable = 0;
    programData.USBVersion = 0x200;
    
    // Cbus settings for certain pins
    programData.Cbus4H = FT_232H_CBUS_TXRXLED;
    programData.Cbus5H = FT_232H_CBUS_IOMODE;
    programData.Cbus6H = FT_232H_CBUS_IOMODE;
    programData.Cbus8H = FT_232H_CBUS_IOMODE;
    programData.Cbus9H = FT_232H_CBUS_IOMODE;

    // FT1248 settings
    programData.IsFT1248H = 1;          // using FT1248 interface
    programData.FT1248CpolH = 1;        // clock polarity idle high
    programData.FT1248LsbH = 0;         // MSB first
    programData.FT1248FlowControlH = 0;
    programData.IsVCPH = 0;             // Using DLL drivers, not VCPa
    
    ftStatus = FT_EE_Program(ftHandle, &programData);
    statusAssert("Did not program EEProm correctly.");
    printf("EEProm programmed successfully\n");
}

void FtdiDevice::clear()
{
    dataBuffer.clearN(dataBuffer.getEntries());
    for(uint8_t i = 0; i < numChannels; ++i)
    {
        channelBuffer[i].setSize(0);
    }
    clearBuffers();
}

void FtdiDevice::openDevice()
{
    ftStatus = FT_OpenEx("UM232H", FT_OPEN_BY_DESCRIPTION, &ftHandle);
    statusAssert("Did not open UM232H device.");
    printf("Opened device\n");
}

void FtdiDevice::closeDevice()
{
    ftStatus = FT_Close(ftHandle);
    statusAssert("Did not close UM232H device.");
    printf("Closed device\n");
}

void FtdiDevice::resetDevice()
{
    ftStatus = FT_SetBitMode(ftHandle, 0x00, FT_BITMODE_RESET);
    statusAssert("Did not reset UM232H device.");

    ftStatus = FT_SetTimeouts(ftHandle, 5000, 1000);
    statusAssert("Did not set timeouts properly.");

    printf("Reset device\n");
}

void FtdiDevice::clearBuffers()
{
    ftStatus = FT_Purge(ftHandle, FT_PURGE_TX | FT_PURGE_RX);
    statusAssert("Did not clear Tx and Rx buffers.");
}

void FtdiDevice::read(uint32_t numSamples)
{
    // clear data buffer
    dataBuffer.clearN(dataBuffer.getEntries());
    for(uint32_t i = 0; i < numChannels; ++i)
    {
        channelBuffer[i].setSize(numSamples + BYTES_TO_READ/64 + 5);
    }

    dataBuffer.setSize((numSamples + 100) * 64);

    readRxBuffer(BYTES_TO_READ);
    // If we had gotten our read to work, we would have uncommented
    // the following lines of code: 
    
    // alignToNextLRCK(HIGH, 32);
    // alignToNextLRCK(LOW, 32);

    // while(dataBuffer.getEntries() < numSamples*64)
    // {
    //     readRxBuffer(BYTES_TO_READ);
    // }

    // while(formatData()) ;
}

void FtdiDevice::readRxBuffer(DWORD RxBytes)
{
    // This API call will block until RxBytes have been read,
    // or until an error occurs

    ftStatus = FT_Read(ftHandle, &RxBuffer, RxBytes, &BytesReceived);
    statusAssert("Did not read Rx buffer.");

    if(BytesReceived < RxBytes)
    {
        printf("Bytes received is: %d, should have been: %d\n", 
                BytesReceived, RxBytes);
    }
    dataBuffer.addN(RxBuffer, (uint32_t) BytesReceived);

    FILE * fp;
    fp = fopen("dataBuffer.txt", "w");
    for(int i = 0; i < 32; ++i)
    {
        fprintf(fp, "dataBuffer[%d] is: %d\n", i, dataBuffer[i]);
    }
    fclose(fp);
    std::this_thread::sleep_for(std::chrono::milliseconds(2000));
}

bool FtdiDevice::formatData()
{
    // Make sure enough data is in buffer
    if(dataBuffer.getEntries() < 50) 
    {
        return false;
    }

    // Save next 64 entries for debugging in case over 32 clocks seen for one sample
    // dataBuffer.getN(history, 64);

    uint8_t entry = 0;
    // Peek into first entry for the LRCK bit, which is high in LJ mode
    // when the channel is odd
    dataBuffer.getN(&entry, 1);
    uint8_t LRCK = entry & 1;

    uint32_t i0 = 0;
    uint32_t i1 = 0;
    uint32_t i2 = 0;
    uint32_t i3 = 0;
    // Add a new entry for the even or odd channels, depending on LRCK
    // and how many channels needed
    if(numChannels >= (2-LRCK)) 
    {
        i0 = channelBuffer[1-LRCK].add(0);
    }
    
    if(numChannels >= (4-LRCK)) 
    {
        i1 = channelBuffer[3-LRCK].add(0);
    }
    
    if(numChannels >= (6-LRCK)) 
    {
        i2 = channelBuffer[5-LRCK].add(0);
    }
    
    if(numChannels >= (8-LRCK))
    {
        i3 = channelBuffer[7-LRCK].add(0);
    }

    // For each of the next 24 bits, store the bit into the correct
    // position of each channel sample buffer
    for(int i = 0; i < 24; ++i)
    {
        dataBuffer.getN(&entry, 1);                 // Get one entry
        
        if(LRCK != (entry&1)) return true;
        
        if(numChannels >= (2-LRCK))
        {
            channelBuffer[1-LRCK][i0] |= ((uint32_t) ((entry&2)>>1) << (23-i));
        }
        
        if(numChannels >= (4-LRCK))
        {
            channelBuffer[3-LRCK][i1] |= ((uint32_t) ((entry&4)>>2) << (23-i));
        }
        
        if(numChannels >= (6-LRCK))
        {
            channelBuffer[5-LRCK][i2] |= ((uint32_t) ((entry&8)>>3) << (23-i));
        }
        
        if(numChannels >= (8-LRCK))
        {
            channelBuffer[7-LRCK][i3] |= ((uint32_t) ((entry&16)>>4) << (23-i));
        }
        
        dataBuffer.clearN(1);                       // Remove one entry
    }

    // Skip extra bits to the next LRCK
    PinState LRCLK = LRCK == 1 ? HIGH : LOW;
    alignToNextLRCK(LRCLK, 8);

    return true;
}

/*
 * When LRCK is HIGH, odd channels are read (1,3,5,7)
 * When LRCK is LOW, even channels are read (2,4,6,8)
*/
void FtdiDevice::alignToNextLRCK(PinState LRCK, uint32_t limit)
{
    uint8_t entry;
    int i = 0;
    // Clear entries from buffer until the entry
    // has a different LRCK value
    while(dataBuffer.getN(&entry, 1) && LRCK == (entry&1))
    {
        dataBuffer.clearN(1);
        ++i;
        if(i > limit)
        {
           printf("Warning: Over 32 clock cycles seen for one sample!\n");
        }
    }
}

void FtdiDevice::setNumChannels(uint8_t channels)
{
    numChannels = channels;
}

uint8_t FtdiDevice::getNumChannels()
{
    return numChannels;
}

void FtdiDevice::statusAssert(const char * message)
{
    if(!FT_SUCCESS(ftStatus))
    {
        printf(message);
        printf("\nftStatus is: %d\n", ftStatus);
        // exit(0);
    }
    return;
}

GpioPin::GpioPin(FtdiDevice* device, uint8_t pin_index) 
    :   ft(device),
        pin_state(LOW),
        mask_high(1 << pin_index),
        mask_low(~(1 << pin_index))
{}

GpioPin::~GpioPin() {}

void GpioPin::set(PinState state)
{
    pin_state = state;
    if(pin_state == LOW)
    {
        ft->cbusState &= mask_low;
    }
    else
    {
        ft->cbusState |= mask_high;
    }
}

void GpioPin::write()
{
    ft->ftStatus = FT_SetBitMode(ft->ftHandle, ft->cbusState, FT_BITMODE_CBUS_BITBANG);
    if(ft->ftStatus)
    {
        printf("GpioPin write had an error. ftStatus is: %d\n", ft->ftStatus);
    }
}

void GpioPin::operator=(PinState state)
{
    set(state);
    write();
}

PinState GpioPin::operator!() const
{
    if(pin_state)
    {
        return LOW;
    }
    return HIGH;
}
