#ifndef USB_INTERFACE_H_
#define USB_INTERFACE_H_

// C++ includes
#include <iostream>
#include <cstdint>
#include <stdio.h>
#include <vector>

// USB Driver Header Include
#include "ftd2xx.h"

// Templated buffer include
#include "buffer.h"

#define DEFAULT_SAMPLING_RATE      192000
#define DEFAULT_NUM_CHANNELS       8

typedef enum
{
    LOW = 0,
    HIGH
} PinState;

class FtdiDevice;

class GpioPin
{
    public:
        GpioPin(FtdiDevice* device, uint8_t pin_index);
        ~GpioPin();

        void set(PinState state);
        void write();

        // set and write
        void operator=(PinState state);
        // return opposite of current state
        PinState operator!() const;

    private:
        FtdiDevice* ft;
        PinState pin_state;
        uint8_t mask_high;
        uint8_t mask_low;
};

class FtdiDevice
{
    friend class GpioPin;
    public:
        FtdiDevice();
        ~FtdiDevice();
        void initDevice();

        void start();
        void clear();
        void read(uint32_t numSamples);
        
        void setNumChannels(uint8_t channels);
        uint8_t getNumChannels();
        
        void setSamplingRate(uint32_t sample_rate);
        uint32_t getSamplingRate();


        GpioPin RST;           // AC5 -- Reset
        GpioPin CDIN;          // AC6 -- Data In
        GpioPin CCLK;          // AC8 -- Clock
        GpioPin CSn;           // AC9 -- Chip Select (active low)

    private:
        void openDevice();
        void closeDevice();
        void resetDevice();
        void clearBuffers();
        void readRxBuffer(DWORD RxBytes);

        void statusAssert(const char * message);

        void programEEPROM();

        bool formatData();
        void alignToNextLRCK(PinState LRCK, uint32_t limit);


        FT_STATUS ftStatus;
        FT_HANDLE ftHandle;
        uint8_t numChannels;

        DWORD BytesReceived;
        uint8_t RxBuffer[256*1024]; // RxBuffer is 256kB
        Buffer<uint8_t> dataBuffer;
        Buffer<uint32_t> channelBuffer[DEFAULT_NUM_CHANNELS];

        /* According to D2XX_Programmer's_Guide:
        *
        * In the case of CBUS Bit Bang, the upper nibble of this value 
        * controls which pins are inputs and outputs,
        * while the lower nibble controls which of the outputs are high and low.
        * 
        * bit 0 = ACBUS 5
        * bit 1 = ACBUS 6
        * bit 2 = ACBUS 8
        * bit 3 = ACBUS 9
        * 
        * This variable holds the information passed to SetBitmode when we set
        * the mode to be CBUS Bit Bang
        */
        uint8_t cbusState;     // maintains current CBUS I/O pin state
        uint8_t chipAddress;   // ADC SPI address
};


#endif //  USB_INTERFACE_H_